package org.hibernate.event.internal;

import java.util.Map;

import org.hibernate.engine.internal.EntityKeyUtil;
import org.hibernate.engine.spi.EntityEntry;
import org.hibernate.engine.spi.EntityKey;
import org.hibernate.engine.spi.PersistenceContext;
import org.hibernate.engine.spi.SessionImplementor;
import org.hibernate.engine.spi.Status;
import org.hibernate.event.spi.LoadEvent;
import org.hibernate.event.spi.LoadEventListener;
import org.hibernate.internal.CoreMessageLogger;
import org.hibernate.internal.SessionExtraInfoHolder;
import org.hibernate.internal.SessionExtraInfoHolderImpl;
import org.hibernate.internal.SessionImplementorWithEntityKeyArray;
import org.hibernate.persister.entity.EntityPersister;
import org.hibernate.pretty.MessageHelper;
import org.hibernate.proxy.HibernateProxy;
import org.hibernate.proxy.LazyInitializer;
import org.jboss.logging.Logger;

/**
 * customize createProxyIfNecessary to not to return concrete entity class, but always retuens proxy
 * @author yding
 *
 */
public class DefaultLoadEventListener2 extends DefaultLoadEventListener {
	 private static final CoreMessageLogger LOG = Logger.getMessageLogger(CoreMessageLogger.class,
             DefaultLoadEventListener2.class.getName());
	 
	 // override to provide customized createProxyIfNecessary
	@Override
	protected Object proxyOrLoad(
		final LoadEvent event,
		final EntityPersister persister,
		final EntityKey keyToLoad,
		final LoadEventListener.LoadType options) {

		if ( LOG.isTraceEnabled() ) {
			LOG.tracev( "Loading entity: {0}",
					MessageHelper.infoString( persister, event.getEntityId(), event.getSession().getFactory() ) );
		}

        // this class has no proxies (so do a shortcut)
        if (!persister.hasProxy()) {
			return load(event, persister, keyToLoad, options);
		}

        final PersistenceContext persistenceContext = event.getSession().getPersistenceContext();

		// look for a proxy
        Object proxy = persistenceContext.getProxy(keyToLoad);
        if (proxy != null) {
			return returnNarrowedProxy(event, persister, keyToLoad, options, persistenceContext, proxy);
		}

        if (options.isAllowProxyCreation()) {
			return createProxyIfNecessary(event, persister, keyToLoad, options, persistenceContext);
		}

        // return a newly loaded object
        return load(event, persister, keyToLoad, options);
	}
	
	// carry over from parent class - no change
	private Object returnNarrowedProxy(
			final LoadEvent event,
			final EntityPersister persister,
			final EntityKey keyToLoad,
			final LoadEventListener.LoadType options,
			final PersistenceContext persistenceContext,
			final Object proxy) {
		LOG.trace( "Entity proxy found in session cache" );
		LazyInitializer li = ( (HibernateProxy) proxy ).getHibernateLazyInitializer();
		if ( li.isUnwrap() ) {
			return li.getImplementation();
		}
		Object impl = null;
		if ( !options.isAllowProxyCreation() ) {
			impl = load( event, persister, keyToLoad, options );
			if ( impl == null ) {
				event.getSession().getFactory().getEntityNotFoundDelegate().handleEntityNotFound( persister.getEntityName(), keyToLoad.getIdentifier());
			}
		}
		return persistenceContext.narrowProxy( proxy, persister, keyToLoad, impl );
	}
	
	
	private Object createProxyIfNecessary(
			final LoadEvent event,
			final EntityPersister persister,
			final EntityKey keyToLoad,
			final LoadEventListener.LoadType options,
			final PersistenceContext persistenceContext) {
		Object existing = persistenceContext.getEntity( keyToLoad );
		if ( existing != null ) {
			// return existing object or initialized proxy (unless deleted)
			LOG.trace( "Entity found in session cache" );
			EntityEntry entry = persistenceContext.getEntry( existing );
			if ( options.isCheckDeleted() ) {
				
				Status status = entry.getStatus();
				if ( status == Status.DELETED || status == Status.GONE ) {
					return null;
				}
			}
			// if its lazily loaded, need to create proxy, otherwise return it directly
			//Map map = persistenceContext.getEntitiesByKey();
			boolean needToCreateProxy = false;
			SessionExtraInfoHolder holder = new SessionExtraInfoHolderImpl();
			EntityKey[] keys = holder.getSessionExtraInfo().getEntityKeys();
			if (!EntityKeyUtil.entityClassInArray(keys, keyToLoad)) {
				needToCreateProxy = true;
			}
			if (!needToCreateProxy) {
				return existing;
			}
			
			
		}
		LOG.trace( "Creating new proxy for entity" );
		// return new uninitialized proxy
		Object proxy = persister.createProxy( event.getEntityId(), event.getSession() );
		persistenceContext.getBatchFetchQueue().addBatchLoadableEntityKey( keyToLoad );
		persistenceContext.addProxy( keyToLoad, proxy ); 
		return proxy;
	}
	
	
}
